"use strict";

exports.__esModule = true;

var _ArrayProperty = require("./ArrayProperty");

var _ArrayProperty2 = _interopRequireDefault(_ArrayProperty);

var _CollectionProperty = require("./collection/CollectionProperty");

var _CollectionProperty2 = _interopRequireDefault(_CollectionProperty);

var _IndexedProperty = require("./IndexedProperty");

var _IndexedProperty2 = _interopRequireDefault(_IndexedProperty);

var _KeyedProperty = require("./KeyedProperty");

var _KeyedProperty2 = _interopRequireDefault(_KeyedProperty);

var _MapProperty = require("./MapProperty");

var _MapProperty2 = _interopRequireDefault(_MapProperty);

var _PropertyFactoryShader = require("./PropertyFactoryShader");

var _PropertyFactoryShader2 = _interopRequireDefault(_PropertyFactoryShader);

var _PrimitiveProperty = require("./PrimitiveProperty");

var _PrimitiveProperty2 = _interopRequireDefault(_PrimitiveProperty);

var _Property = require("./Property");

var _Property2 = _interopRequireDefault(_Property);

var _StateType = require("./StateType");

var _StateType2 = _interopRequireDefault(_StateType);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.default = {
	get Primitive() {
		return new _StateType2.default(_PrimitiveProperty2.default);
	},

	get Array() {
		var type = new _StateType2.default(_ArrayProperty2.default);

		return type;
	},

	arrayOf: function arrayOf(elementStateType) {
		var type = new _StateType2.default(_ArrayProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},


	get Indexed() {
		var type = new _StateType2.default(_IndexedProperty2.default);

		return type;
	},

	indexedOf: function indexedOf(elementStateType) {
		var type = new _StateType2.default(_IndexedProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},


	// CollectionProperty needs refactoring to support
	// collectionOf(elementStateType) {
	// 	const type = new StateType(CollectionProperty);

	// 	// will call the setElementType() method after property created - will need to add
	// 	// functionality to the factory shader
	// 	type.setManagedPropertyType(elementStateType);

	// 	return type;
	// },

	keyed: function keyed() {
		var defn = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

		var type = new _StateType2.default(_KeyedProperty2.default);
		var propType;

		for (var key in defn) {
			propType = defn[key];

			type.addProperty(key, defn[key]);
		}

		return type;
	},


	// Deprecated - use MapProperty.mapType()
	map: function map() {
		var defn = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

		var type = new _StateType2.default(_MapProperty2.default);
		var propType;

		for (var key in defn) {
			propType = defn[key];

			type.addProperty(key, defn[key]);
		}

		return type;
	},


	// Deprecated - use MapProperty.mapTypeOf()
	mapOf: function mapOf(elementStateType) {
		var type = new _StateType2.default(_MapProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},
	property: function property(PropertyClass) {
		if (PropertyClass.stateSpec) {
			return PropertyClass.stateSpec;
		}

		return new _StateType2.default(PropertyClass);
	}
};