"use strict";

exports.__esModule = true;
exports.default = createObjectPropertyType;

var _ObjectProperty2 = require("./ObjectProperty");

var _ObjectProperty3 = _interopRequireDefault(_ObjectProperty2);

var _SimpleShadow2 = require("./SimpleShadow");

var _SimpleShadow3 = _interopRequireDefault(_SimpleShadow2);

var _Shadow = require("./Shadow");

var _Shadow2 = _interopRequireDefault(_Shadow);

var _StateTypes = require("./StateTypes");

var _StateTypes2 = _interopRequireDefault(_StateTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/*
	Creates an ObjectProperty subclass based on a custom Shadow type.
*/
function createObjectPropertyType() {
	var shadowType = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
	var stateSpec = arguments[1];
	var specCallback = arguments[2];

	var ShadowClass;

	// get the shadow class
	if (shadowType instanceof _SimpleShadow3.default || shadowType instanceof _Shadow2.default) {
		// shadow class passed into method
		ShadowClass = shadowType;
	} else {
		var CustomObjectShadow = function (_SimpleShadow) {
			_inherits(CustomObjectShadow, _SimpleShadow);

			function CustomObjectShadow() {
				_classCallCheck(this, CustomObjectShadow);

				return _possibleConstructorReturn(this, Object.getPrototypeOf(CustomObjectShadow).apply(this, arguments));
			}

			return CustomObjectShadow;
		}(_SimpleShadow3.default);

		var proto = CustomObjectShadow.prototype;
		var names = Object.getOwnPropertyNames(shadowType);
		var name, desc;

		for (var i = 0, len = names.length; i < len; i++) {
			name = names[i];
			desc = Object.getOwnPropertyDescriptor(shadowType, name);

			Object.defineProperty(proto, name, desc);
		}

		ShadowClass = CustomObjectShadow;
	}

	// default spec for constructor
	var spec = stateSpec || {};

	// create the property subclass

	var CustomObjectProperty = function (_ObjectProperty) {
		_inherits(CustomObjectProperty, _ObjectProperty);

		function CustomObjectProperty() {
			var initialState = arguments.length <= 0 || arguments[0] === undefined ? spec._initialState : arguments[0];
			var autoShadow = arguments.length <= 1 || arguments[1] === undefined ? spec._autoshadow : arguments[1];
			var readonly = arguments.length <= 2 || arguments[2] === undefined ? spec._readonly : arguments[2];

			_classCallCheck(this, CustomObjectProperty);

			var _this2 = _possibleConstructorReturn(this, Object.getPrototypeOf(CustomObjectProperty).call(this, initialState, autoShadow, readonly));

			_this2.setShadowClass(ShadowClass);
			return _this2;
		}

		return CustomObjectProperty;
	}(_ObjectProperty3.default);

	// assign state spec if present to new Property subclass


	CustomObjectProperty.stateSpec = stateSpec || _StateTypes2.default.property(CustomObjectProperty);

	if (specCallback) {
		specCallback(CustomObjectProperty.stateSpec);
	}

	return CustomObjectProperty;
}